﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using System.IO;
using System.Xml;
using System.Globalization;

namespace Microsoft.Tools.WindowsInstallerXml.Build.Tasks
{
    /// <summary>
    /// This task assigns Culture metadata to files based on the value of the Culture attribute on the
    /// WixLocalization element inside the file.
    /// </summary>
    public class WixAssignCulture : Task
    {
        private const string CultureAttributeName = "Culture";
        private const string CultureMetadataName = "Culture";

        private ITaskItem[] files;
        private List<ITaskItem> assignedFiles = new List<ITaskItem>();
        private List<ITaskItem> skippedFilesWithNoCulture = new List<ITaskItem>();
        private List<ITaskItem> skippedFilesWithExistingCulture = new List<ITaskItem>();
        private List<ITaskItem> allFilesWithCulture = new List<ITaskItem>();

        /// <summary>
        /// The list of files to apply culture information to.
        /// </summary>
        [Required]
        public ITaskItem[] Files
        {
            get { return this.files; }
            set { this.files = value; }
        }

        /// <summary>
        /// The files that had culture information applied
        /// </summary>
        [Output]
        public ITaskItem[] AllFilesWithCulture
        {
            get { return this.assignedFiles.ToArray(); }
        }

        /// <summary>
        /// Applies culture information to the files specified by the Files property.
        /// </summary>
        /// <returns>True upon completion of the task execution.</returns>
        public override bool Execute()
        {
            foreach (ITaskItem file in this.files)
            {
                XmlDocument wxlFile = new XmlDocument();

                try
                {
                    wxlFile.Load(file.ItemSpec);
                }
                catch (FileNotFoundException)
                {
                    Log.LogError("Unable to retrieve the culture for file {0}. The file was not found.", file.ItemSpec);
                    return false;
                }
                catch (Exception e)
                {
                    Log.LogError("Unable to retrieve the culture for file {0}: {1}", file.ItemSpec, e.Message);
                    return false;
                }

                // Check and see if the attribute is there. The XSD requires it, but this task runs before candle has a chance to
                // do validation on the file.
                if (wxlFile.DocumentElement.Attributes[WixAssignCulture.CultureAttributeName] == null)
                {
                    Log.LogError("File {0} has no Culture attribute on the WixLocalization element. Add the attribute to specify the file's culture then try building again.", file.ItemSpec);
                    return false;
                }

                // Take the culture value and try using it to create a culture.
                string culture = wxlFile.DocumentElement.Attributes[WixAssignCulture.CultureAttributeName].Value;

                try
                {
                    CultureInfo.GetCultureInfo(culture);
                }
                catch (ArgumentException)
                {
                    Log.LogError("The culture {0} specified for file {0} is not a valid culture name. Modify the culture name on the Culture attribute of the WixLocalization element then try building again.", culture, file.ItemSpec);
                    return false;
                }

                // At this point we have a valid culture so apply it as metadata
                Log.LogMessage(MessageImportance.Low, "Applied culture \"{0}\" to file {1}.", culture, file.ItemSpec);
                file.SetMetadata(WixAssignCulture.CultureMetadataName, culture);
                this.assignedFiles.Add(file);
            }

            return true;
        }
    }
}
